/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2023 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include "config.h"

#include <librnd/config.h>
#include <librnd/core/actions.h>
#include <librnd/hid/hid.h>

#include "build_run.h"
#include "conf_core.h"
#include "data.h"
#include "obj_any.h"
#include "search.h"
#include "projection.h"
#include "funchash_core.h"

static int merge_nodes(osm_layer_t *ly, vtp0_t *lst)
{
	long n, m;
	double cx = 0, cy = 0, lat, lon;
	htdr_t dr = {0};
	osm_rtree_box_t bbox;
	rnd_coord_t ccx, ccy;
	osm_node_t *nd;

	bbox.x2 = bbox.y2 = RND_COORD_MAX;
	bbox.x1 = bbox.y1 = -RND_COORD_MAX;
	htdr_init(&dr, htdr_keyhash, htdr_keyeq);

	for(n = 0; n < lst->used; n++) {
		nd = lst->array[n];
		cx += nd->x;
		cy += nd->y;
		osm_rtree_box_bump(&bbox, osm_obj_bbox((osm_any_obj_t *)nd));
		for(m = n; m < lst->used; m++) {
			osm_node_t *nd2 = lst->array[m];
			if (osm_tags_conflict(&nd->curr->tags, &nd2->curr->tags)) {
				rnd_message(RND_MSG_ERROR, "Tag conflict between nodes " OSM_PID " and " OSM_PID "\n", nd->hdr.id, nd2->hdr.id);
				return -1;
			}
		}
	}

	ccx = rnd_round(cx/lst->used);
	ccy = rnd_round(cy/lst->used);

	uundo_freeze_serial(&ly->undo);

	nd = lst->array[0];
	osm_node_pre_edit((osm_any_obj_t *)nd, 1, 1);

	rnd2lldeg(&lon, &lat, ccx, ccy);
	osm_node_modify(ly, nd, &lon, &lat, 1);

	for(n = 1; n < lst->used; n++) {
		osm_node_t *nd2 = lst->array[n];
		htdr_key_t drk;
		htdr_value_t drv;

		drk.type = OSM_OBJ_NODE;
		drk.old_id = nd2->hdr.id;
		drv.valid = 1;
		drv.new_id = nd->hdr.id;
		drv.new_version = nd->hdr.version;
		htdr_set(&dr, drk, drv);

		osm_tags_merge(&nd->curr->tags, &nd2->curr->tags);
	}
	osm_node_post_edit((osm_any_obj_t *)nd, 1);

	/* Update al referers within the given bbox */
	{
		osm_rtree_it_t it;
		vtp0_t orphaned = {0};
		void *o;
		long n;

		for(o = osm_rtree_first(&it, &ly->objs, &bbox); o != NULL; o = osm_rtree_next(&it)) {
			osm_any_obj_t *obj = o;

			if (obj->hdr.type == OSM_OBJ_NODE) continue; /* otimization: nodes don't refer to other objects */
			obj->hdr.calls->diffresult_referees(obj, &dr, 0, 1);

			/* schedule removal of objects that got empty */
			if ((obj->hdr.type == OSM_OBJ_WAY) && (obj->way.curr->nd.used < 2)) {
				if (!obj->hdr.referer) {
					vtp0_append(&orphaned, obj);
					obj->hdr.referer = 1;
				}
			}
		}

		/* remove orphaned objects; can't be done during rtree search because
		   it changes the rtree */
		for(n = 0; n < orphaned.used; n++) {
			osm_any_obj_t *obj = orphaned.array[n];
			obj->hdr.referer = 0;
			osm_obj_remove(obj, 1);
		}

		vtp0_uninit(&orphaned);
	}

	for(n = 1; n < lst->used; n++)
		osm_obj_remove(lst->array[n], 1);

	uundo_unfreeze_serial(&ly->undo);
	uundo_inc_serial(&ly->undo);

	htdr_uninit(&dr);
	return 0;
}

static int shared_way_ept(osm_layer_t *ly, osm_way_t *w1, int end1, osm_way_t *w2, int end2, int *twin)
{
	osm_id_t ni1, ni2;
	osm_node_t *n1, *n2;
	double dx, dy, dist;

	*twin = 0;

	ni1 = end1 ? w1->curr->nd.array[w1->curr->nd.used-1] : w1->curr->nd.array[0];
	ni2 = end2 ? w2->curr->nd.array[w2->curr->nd.used-1] : w2->curr->nd.array[0];

	if (ni1 == ni2)
		return 1;

	n1 = (osm_node_t *)osm_obj_by_id(ly, OSM_OBJ_NODE, ni1);
	if (n1 == NULL)
		return 0;
	n2 = (osm_node_t *)osm_obj_by_id(ly, OSM_OBJ_NODE, ni2);
	if (n2 == NULL)
		return 0;

	dx = n1->x - n2->x;
	dy = n1->y - n2->y;
	dist = (dx * dx + dy * dy) / 10000.0;

	*twin = (dist <= conf_core.editor.merge_way_endp_dist2);
	return *twin;
}

/* Append nodes of src at the end of dst */
static void way_append(osm_way_t *dst, osm_way_t *src, int twin)
{
	long n, start;

	if (twin)
		start = 1; /* we have two separate nodes but they are very close; skip src's start node */
	else if (dst->curr->nd.array[dst->curr->nd.used-1] == src->curr->nd.array[0])
		start = 1; /* don't append the same node */
	else
		start = 0;

	for(n = start; n < src->curr->nd.used; n++)
		osm_way_modify_append(dst->hdr.layer, dst, src->curr->nd.array[n], 1);
}

static int merge_ways(osm_layer_t *ly, vtp0_t *lst)
{
	long n1, n2;
	int done = 0;

	uundo_freeze_serial(&ly->undo);

	restart:;

	TODO("Implement the details");

	/* pick a pair that share endpoint */
	for(n1 = 0; n1 < lst->used; n1++) {
		osm_way_t *w1 = lst->array[n1];
		for(n2 = n1+1; n2 < lst->used; n2++) {
			osm_way_t *w2 = lst->array[n2];
			int twin;

			if (shared_way_ept(ly, w1, 0, w2, 0, &twin)) {
				/* reverse w1 and append w2 to w1 */
				osm_way_modify_reverse(w1, 1);
				way_append(w1, w2, twin);
				vtp0_remove(lst, n2, 1); /* del w2, w1 remains */
				osm_way_pre_edit(w1, 0, 1);
				osm_tags_merge(&w1->curr->tags, &w2->curr->tags);
				osm_way_post_edit(w1, 1);
				osm_obj_remove((osm_any_obj_t *)w2, 1);
				done = 1;
				goto restart;
			}
			if (shared_way_ept(ly, w1, 1, w2, 0, &twin)) {
				/* append w2 to w1 */
				way_append(w1, w2, twin);
				vtp0_remove(lst, n2, 1); /* del w2, w1 remains */
				osm_way_pre_edit(w1, 0, 1);
				osm_tags_merge(&w1->curr->tags, &w2->curr->tags);
				osm_way_post_edit(w1, 1);
				osm_obj_remove((osm_any_obj_t *)w2, 1);
				done = 1;
				goto restart;
			}
			if (shared_way_ept(ly, w1, 1, w2, 1, &twin)) {
				/* reverse w2 and append w2 to w1 */
				osm_way_modify_reverse(w2, 1);
				way_append(w1, w2, twin);
				vtp0_remove(lst, n2, 1); /* del w2, w1 remains */
				osm_way_pre_edit(w1, 0, 1);
				osm_tags_merge(&w1->curr->tags, &w2->curr->tags);
				osm_way_post_edit(w1, 1);
				osm_obj_remove((osm_any_obj_t *)w2, 1);
				done = 1;
				goto restart;
			}
			if (shared_way_ept(ly, w1, 0, w2, 1, &twin)) {
				/* append w1 to w2 */
				way_append(w2, w1, twin);
				vtp0_remove(lst, n1, 1); /* del w1, w2 remains */
				osm_way_pre_edit(w2, 0, 1);
				osm_tags_merge(&w2->curr->tags, &w1->curr->tags);
				osm_way_post_edit(w2, 1);
				osm_obj_remove((osm_any_obj_t *)w1, 1);
				done = 1;
				goto restart;
			}
		}
	}

	if (!done)
		rnd_message(RND_MSG_ERROR, "Merging ways failed: no matched endpoints\n");

	uundo_unfreeze_serial(&ly->undo);
	uundo_inc_serial(&ly->undo);

	return 0;
}

static const char osm_acts_Merge[] = "Merge()";
static const char osm_acth_Merge[] = "Merges selected nodes or selected ways.";
static fgw_error_t osm_act_Merge(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	osm_design_t *osm = (osm_design_t *)RND_ACT_DESIGN;
	osm_layer_t *ly = osm_layer_get_curr(osm);
	vtp0_t lst = {0};
	long n, nnodes = 0, nways = 0;

	if (ly == NULL) {
		rnd_message(RND_MSG_ERROR, "No active layer");
		return -1;
	}

	osm_search_all_selected(&lst, ly);

	for(n = 0; n < lst.used; n++) {
		osm_any_obj_t *obj = lst.array[n];
		if (obj->hdr.type == OSM_OBJ_NODE)
			nnodes++;
		else if (obj->hdr.type == OSM_OBJ_WAY)
			nways++;
		if ((nnodes != 0) && (nways != 0)) {
			rnd_message(RND_MSG_ERROR, "Need to select only ways or only nodes\n");
			goto error;
		}
	}

	if (nnodes + nways < 2) {
		rnd_message(RND_MSG_ERROR, "Too few objects are selected\n");
		goto error;
	}

	if (nnodes > 0) {
		if (merge_nodes(ly, &lst) < 0)
			goto error;
	}
	else if (nways > 0) { /* merge ways */
		if (merge_ways(ly, &lst) < 0)
			goto error;
	}

	vtp0_uninit(&lst);
	return FGW_SUCCESS;

	error:;
	vtp0_uninit(&lst);
	return -1;
}

static const char osm_acts_Reverse[] = "Reverse(object|selected)";
static const char osm_acth_Reverse[] = "Reverses the direction of addressed way(s).";
static fgw_error_t osm_act_Reverse(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	osm_design_t *osm = (osm_design_t *)RND_ACT_DESIGN;
	osm_layer_t *ly = osm_layer_get_curr(osm);
	int op;

	RND_ACT_CONVARG(1, FGW_KEYWORD, Reverse, op = fgw_keyword(&argv[1]));

	if (ly == NULL) {
		rnd_message(RND_MSG_ERROR, "No active layer");
		return -1;
	}

	switch(op) {
		case F_Object:
		{
			osm_any_obj_t *way = osm_search_obj_xy_mask(ly, osm->crosshair_x, osm->crosshair_y, OSM_MASK_WAY);
			osm_way_modify_reverse(&way->way, 1);
			break;
		}

		case F_Selected:
		{
			vtp0_t sw = {0};
			long n;

			osm_search_all_selected_mask(&sw, ly, OSM_MASK_WAY);

			uundo_freeze_serial(&ly->undo);

			for(n = 0; n < sw.used; n++)
				osm_way_modify_reverse(sw.array[n], 1);

			uundo_unfreeze_serial(&ly->undo);
			uundo_inc_serial(&ly->undo);

			vtp0_uninit(&sw);
		}
	}

	return 0;
}

/* Argv is already all-strings */
static int tagset(osm_any_obj_t *obj, int argc, fgw_arg_t *argv)
{
	int n, changed = 0;

	if (obj == NULL)
		return 0;

	for(n = 0; n < argc; n++) {
		char *key = argv[n].val.str, *val = n < argc-1 ? argv[n+1].val.str : "";
		if (*key == '@') {
			int op = rnd_funchash_get(key+1, NULL);
			switch(op) {
				case F_Closed:
					if (obj->hdr.type != OSM_OBJ_WAY)
						goto quit;
					if (obj->way.curr->nd.used < 3)
						goto quit;
					if (obj->way.curr->nd.array[0] != obj->way.curr->nd.array[obj->way.curr->nd.used-1])
						goto quit;
					break;
				case F_Node:
					if (obj->hdr.type != OSM_OBJ_NODE)
						goto quit;
					break;
				case F_Way:
					if (obj->hdr.type != OSM_OBJ_WAY)
						goto quit;
					break;
				case F_Relation:
					if (obj->hdr.type != OSM_OBJ_RELATION)
						goto quit;
					break;
			}
		}
		else {
			osm_tags_t *tags;

			if (obj->hdr.calls->get_tags == NULL)
				break;

			if (!changed)
				osm_obj_pre_edit(obj, 0);

			tags = obj->hdr.calls->get_tags(obj);
			if (htsp_has(tags, key))
				osm_obj_modify_tag(obj->hdr.layer, obj, key, val, 1);
			else
				osm_tag_create(obj, key, val, 1);
			n++;
			changed = 1;
		}
	}

	quit:;
	if (changed)
		osm_obj_post_edit(obj, 1);
	return changed;
}

static const char osm_acts_Tagset[] = "Tagset(object|selected|auto, instructions...)";
static const char osm_acth_Tagset[] = "Set tags on object(s).";
static fgw_error_t osm_act_Tagset(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	osm_design_t *osm = (osm_design_t *)RND_ACT_DESIGN;
	osm_layer_t *ly = osm_layer_get_curr(osm);
	int op, n;
	long changed = 0;

	RND_ACT_CONVARG(1, FGW_KEYWORD, Tagset, op = fgw_keyword(&argv[1]));
	for(n = 2; n < argc; n++)
		RND_ACT_CONVARG(n, FGW_STR, Tagset, (void)argv[n].val.str);

	if (ly == NULL) {
		rnd_message(RND_MSG_ERROR, "No active layer");
		return -1;
	}

	switch(op) {
		case F_Object:
		{
			osm_any_obj_t *obj;
			obj1:;
			obj = osm_search_obj_xy(ly, osm->crosshair_x, osm->crosshair_y);
			changed = tagset(obj, argc-2, argv+2);
			break;
		}

		case F_Selected:
		case F_Auto:
		{
			vtp0_t sw = {0};
			long n;

			osm_search_all_selected_mask(&sw, ly, OSM_MASK_WAY);

			if ((sw.used == 0) && (op == F_Auto))
				goto obj1;

			uundo_freeze_serial(&ly->undo);

			for(n = 0; n < sw.used; n++)
				changed += tagset(sw.array[n], argc-2, argv+2);

			uundo_unfreeze_serial(&ly->undo);
			uundo_inc_serial(&ly->undo);

			vtp0_uninit(&sw);
		}
	}

	if (changed == 0)
		rnd_message(RND_MSG_ERROR, "TagSet(): no object qualified, no modification made\n");
	else
		rnd_message(RND_MSG_INFO, "TagSet(): changed %ld objects\n", changed);

	return 0;
}

static rnd_action_t edit_action_list[] = {
	{"Merge", osm_act_Merge, osm_acth_Merge, osm_acts_Merge},
	{"Reverse", osm_act_Reverse, osm_acth_Reverse, osm_acts_Reverse},
	{"Tagset", osm_act_Tagset, osm_acth_Tagset, osm_acts_Tagset}
};

void edit_act_init(void)
{
	RND_REGISTER_ACTIONS(edit_action_list, NULL)
}

