/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

/* included from obj_way.c to split the long code */

#include "undo.h"

static const char obj_way_cookie[] = "way";

typedef struct {
	osm_id_t id;
	osm_layer_t *ly;
} undo_way_fork_t;

static int undo_way_fork_swap(void *udata)
{
	undo_way_fork_t *u = udata;
	osm_way_t *way = htIp_get(&u->ly->id2obj[OSM_OBJ_WAY], u->id);

	if (way == NULL) {
		rnd_message(RND_MSG_ERROR, "undo_way_fork_swap(): way not found: " OSM_PID "\n", u->id);
		return -1;
	}

	if (way->curr == &way->orig)
		osm_way_fork(way);
	else
		osm_way_unfork(way);

	return 0;
}

static void undo_way_fork_print(void *udata, char *dst, size_t dst_len)
{
	undo_way_fork_t *u = udata;
	osm_way_t *way = htIp_get(&u->ly->id2obj[OSM_OBJ_WAY], u->id);

	if (way == NULL) {
		rnd_snprintf(dst, dst_len, "way fork: invalid id " OSM_PID, u->id);
		return;
	}

	rnd_snprintf(dst, dst_len, "way %s", (way->curr == &way->orig) ? "fork" : "unfork");
}


static const uundo_oper_t undo_way_fork = {
	obj_way_cookie,
	NULL,
	undo_way_fork_swap,
	undo_way_fork_swap,
	undo_way_fork_print
};


void osm_way_modify_fork(osm_layer_t *layer, osm_way_t *way, int undoable)
{
	undo_way_fork_t utmp, *u = &utmp;

	if (undoable) u = uundo_append(&layer->undo, &undo_way_fork, sizeof(undo_way_fork_t));

	u->ly = layer;
	u->id = way->hdr.id;

	undo_way_fork_swap(u);
	if (undoable) osm_undo_inc_serial(layer);
}




typedef struct {
	osm_id_t id;
	osm_layer_t *ly;

	osm_id_t node_id;
	long node_idx;
	enum {
		WMOD_ADD,
		WMOD_DEL,
		WMOD_REPLACE,
		WMOD_REVERSE
	} op;
} undo_way_modify_t;

static int undo_way_modify_swap(void *udata)
{
	undo_way_modify_t *u = udata;
	osm_way_t *way = htIp_get(&u->ly->id2obj[OSM_OBJ_WAY], u->id);
	long n;

	if (way == NULL) {
		rnd_message(RND_MSG_ERROR, "undo_way_modify_swap(): way not found: " OSM_PID "\n", u->id);
		return -1;
	}

	TODO("Figure redraw");
/* osm_obj_redraw(way); */


	osm_way_pre_edit((osm_any_obj_t *)way, 1, 1);

	switch(u->op) {
		case WMOD_ADD:
			assert(u->node_idx >= 0);
			assert(u->node_idx <= way->curr->nd.used);
			vtI0_insert_len(&way->curr->nd, u->node_idx, &u->node_id, 1);
			u->op = WMOD_DEL;
			break;

		case WMOD_DEL:
			assert(u->node_idx >= 0);
			assert(u->node_idx < way->curr->nd.used);
			assert(way->curr->nd.array[u->node_idx] == u->node_id);
			vtI0_remove(&way->curr->nd, u->node_idx, 1);
			u->op = WMOD_ADD;
			break;

		case WMOD_REPLACE:
			assert(u->node_idx >= 0);
			assert(u->node_idx <= way->curr->nd.used);
			rnd_swap(osm_id_t, way->curr->nd.array[u->node_idx], u->node_id);
			break;

		case WMOD_REVERSE:
			for(n = 0; n < way->curr->nd.used/2; n++)
				rnd_swap(osm_id_t, way->curr->nd.array[n], way->curr->nd.array[way->curr->nd.used-n-1]);
			break;
	}

	osm_way_post_edit((osm_any_obj_t *)way, 0);

	TODO("Figure layer changed flag");
	TODO("Figure redraw");
/* osm_obj_redraw(way); */

	return 0;
}

static void undo_way_modify_print(void *udata, char *dst, size_t dst_len)
{
	undo_way_modify_t *u = udata;
	osm_way_t *way = htIp_get(&u->ly->id2obj[OSM_OBJ_WAY], u->id);
	char *opname;

	if (way == NULL) {
		rnd_snprintf(dst, dst_len, "way modify: invalid id " OSM_PID, u->id);
		return;
	}

	switch(u->op) {
		case WMOD_ADD:     opname = "add"; break;
		case WMOD_DEL:     opname = "del"; break;
		case WMOD_REPLACE: opname = "replace"; break;
		case WMOD_REVERSE: 
			rnd_snprintf(dst, dst_len, "way reverse #" OSM_PID, u->id);
			return;
	}

	rnd_snprintf(dst, dst_len, "way #%ld %s point #" OSM_PID " at %ld",
		u->id, opname, u->node_id, u->node_idx);
}

static void undo_way_modify_free(void *udata)
{
/*	undo_way_modify_t *u = udata;*/
}

static const uundo_oper_t undo_way_modify = {
	obj_way_cookie,
	undo_way_modify_free,
	undo_way_modify_swap,
	undo_way_modify_swap,
	undo_way_modify_print
};


void osm_way_modify_append(osm_layer_t *layer, osm_way_t *way, osm_id_t node_id, int undoable)
{
	undo_way_modify_t utmp, *u = &utmp;

	if (undoable) {
		uundo_freeze_serial(&layer->undo);

		if (way->curr == &way->orig)
			osm_way_modify_fork(layer, way, undoable);
	}

	if (undoable) u = uundo_append(&layer->undo, &undo_way_modify, sizeof(undo_way_modify_t));

	u->ly = layer;
	u->id = way->hdr.id;
	u->node_id = node_id;
	u->node_idx = way->curr->nd.used;
	u->op = WMOD_ADD;

	undo_way_modify_swap(u);
	if (undoable) osm_undo_inc_serial(layer);
	else undo_way_modify_free(u);

	if (undoable) {
		uundo_unfreeze_serial(&layer->undo);
		uundo_inc_serial(&layer->undo);
	}
}

void osm_way_modify_insert_before(osm_layer_t *layer, osm_way_t *way, osm_id_t node_id, long after_nd_idx, int undoable)
{
	undo_way_modify_t utmp, *u = &utmp;

	if (undoable) {
		uundo_freeze_serial(&layer->undo);

		if (way->curr == &way->orig)
			osm_way_modify_fork(layer, way, undoable);
	}

	if (undoable) u = uundo_append(&layer->undo, &undo_way_modify, sizeof(undo_way_modify_t));

	u->ly = layer;
	u->id = way->hdr.id;
	u->node_id = node_id;
	u->node_idx = after_nd_idx;
	u->op = WMOD_ADD;

	undo_way_modify_swap(u);
	if (undoable) osm_undo_inc_serial(layer);
	else undo_way_modify_free(u);

	if (undoable) {
		uundo_unfreeze_serial(&layer->undo);
		uundo_inc_serial(&layer->undo);
	}
}


void osm_way_modify_remove(osm_way_t *way, long idx, int undoable)
{
	undo_way_modify_t utmp, *u = &utmp;
	osm_layer_t *layer = way->hdr.layer;
	osm_node_t *node;

	osm_way_ensure_ndp_(way);

	assert(idx >= 0);
	assert(idx < way->ndp.used);

	node = way->ndp.array[idx];

	if (undoable) {
		uundo_freeze_serial(&layer->undo);

		if (way->curr == &way->orig)
			osm_way_modify_fork(layer, way, undoable);
	}

	if (undoable) u = uundo_append(&layer->undo, &undo_way_modify, sizeof(undo_way_modify_t));

	u->ly = layer;
	u->id = way->hdr.id;
	u->node_id = node->hdr.id;
	u->node_idx = idx;
	u->op = WMOD_DEL;

	undo_way_modify_swap(u);
	if (undoable) osm_undo_inc_serial(layer);
	else undo_way_modify_free(u);

	if (undoable) {
		uundo_unfreeze_serial(&layer->undo);
		uundo_inc_serial(&layer->undo);
	}
}

void osm_way_modify_replace(osm_way_t *way, long idx, osm_id_t new_node_id, int undoable)
{
	undo_way_modify_t utmp, *u = &utmp;
	osm_layer_t *layer = way->hdr.layer;
	osm_node_t *node;

	osm_way_ensure_ndp_(way);

	assert(idx >= 0);
	assert(idx < way->ndp.used);

	node = way->ndp.array[idx];

	if (undoable) {
		uundo_freeze_serial(&layer->undo);

		if (way->curr == &way->orig)
			osm_way_modify_fork(layer, way, undoable);
	}

	if (undoable) u = uundo_append(&layer->undo, &undo_way_modify, sizeof(undo_way_modify_t));

	u->ly = layer;
	u->id = way->hdr.id;
	u->node_id = new_node_id;
	u->node_idx = idx;
	u->op = WMOD_REPLACE;

	undo_way_modify_swap(u);
	if (undoable) osm_undo_inc_serial(layer);
	else undo_way_modify_free(u);

	if (undoable) {
		uundo_unfreeze_serial(&layer->undo);
		uundo_inc_serial(&layer->undo);
	}
}

void osm_way_modify_reverse(osm_way_t *way, int undoable)
{
	undo_way_modify_t utmp, *u = &utmp;
	osm_layer_t *layer = way->hdr.layer;

	osm_way_ensure_ndp_(way);

	if (undoable) {
		uundo_freeze_serial(&layer->undo);

		if (way->curr == &way->orig)
			osm_way_modify_fork(layer, way, undoable);
	}

	if (undoable) u = uundo_append(&layer->undo, &undo_way_modify, sizeof(undo_way_modify_t));

	u->ly = layer;
	u->id = way->hdr.id;
	u->op = WMOD_REVERSE;

	undo_way_modify_swap(u);
	if (undoable) osm_undo_inc_serial(layer);
	else undo_way_modify_free(u);

	if (undoable) {
		uundo_unfreeze_serial(&layer->undo);
		uundo_inc_serial(&layer->undo);
	}
}

osm_node_t *osm_way_modify_insert_near(osm_way_t *way, rnd_coord_t px, rnd_coord_t py, rnd_coord_t maxdist2, int undoable)
{
	long best_idx;
	rnd_coord_t best_x, best_y;


	best_idx = osm_way_find_near(way, px, py, maxdist2, &best_x, &best_y);

	/* if found the best candidate, create and insert the new node */
	if (best_idx >= 0) {
		osm_layer_t *layer = way->hdr.layer;
		osm_any_obj_t *newnd;

		if (undoable)
			uundo_freeze_serial(&layer->undo);

		newnd = osm_obj_create(layer, OSM_OBJ_NODE, undoable);
		newnd->node.att_x = best_x; newnd->node.att_y = best_y;
		osm_node_move_to_att(&newnd->node, undoable);

		osm_way_modify_insert_before(layer, way, newnd->hdr.id, best_idx, undoable);

		if (undoable) {
			uundo_unfreeze_serial(&layer->undo);
			uundo_inc_serial(&layer->undo);
		}

		return &newnd->node;
	}

	return NULL;
}

static osm_way_t *osm_way_modify_split_at_(osm_way_t *way, long node_idx, int dup_node, int undoable)
{
	osm_layer_t *ly = way->hdr.layer;
	osm_any_obj_t *new_way = NULL;
	long n;

	osm_way_ensure_ndp_(way);

	if (undoable)
		uundo_freeze_serial(&ly->undo);

	new_way = osm_obj_create(ly, OSM_OBJ_WAY, undoable);
	if (new_way == NULL)
		goto error;

	osm_way_pre_edit((osm_any_obj_t *)way, 1, 1);
	osm_way_pre_edit(new_way, 1, 1);

	/* copy over common parts */
	osm_tags_copy(&new_way->way.curr->tags, &way->curr->tags);

	/* Copy over nodes and split */
	if (dup_node) {
		osm_node_t *old_nd = way->ndp.array[node_idx];
		osm_any_obj_t *new_nd = osm_obj_create(ly, OSM_OBJ_NODE, undoable);

		osm_node_pre_edit(new_nd, 1, 1);
		new_nd->node.curr->lat = old_nd->curr->lat;
		new_nd->node.curr->lon = old_nd->curr->lon;
		osm_tags_copy(&new_nd->node.curr->tags, &old_nd->curr->tags);
		osm_node_post_edit(new_nd, 1);
		vtI0_append(&new_way->way.curr->nd, new_nd->hdr.id);
	}
	for(n = node_idx + !!dup_node; n < way->curr->nd.used; n++)
		vtI0_append(&new_way->way.curr->nd, way->curr->nd.array[n]);
	way->curr->nd.used = node_idx+1;

	osm_way_post_edit(new_way, 1);
	osm_way_post_edit((osm_any_obj_t *)way, 0);

	error:;
	if (undoable) {
		uundo_unfreeze_serial(&ly->undo);
		uundo_inc_serial(&ly->undo);
	}
	return &new_way->way;
}

osm_way_t *osm_way_modify_split_at(osm_way_t *way, osm_node_t *node, int dup_node, int undoable)
{
	long node_idx = -1, n;

	osm_way_ensure_ndp_(way);

	for(n = 0; n < way->ndp.used; n++) {
		if (way->ndp.array[n] == node) {
			node_idx = n;
			break;
		}
	}

	if (node_idx < 0)
		return NULL;

	return osm_way_modify_split_at_(way, node_idx, dup_node, undoable);
}

osm_way_t *osm_way_modify_split_near(osm_way_t *way, rnd_coord_t px, rnd_coord_t py, rnd_coord_t maxdist2, int dup_node, int undoable)
{
	osm_layer_t *layer = way->hdr.layer;
	osm_node_t *node;
	osm_way_t *res = NULL;

	if (undoable)
		uundo_freeze_serial(&layer->undo);

	node = osm_way_modify_insert_near(way, px, py, maxdist2, undoable);
	if (node != NULL)
		res = osm_way_modify_split_at(way, node, dup_node, undoable);

	if (undoable) {
		uundo_unfreeze_serial(&layer->undo);
		uundo_inc_serial(&layer->undo);
	}

	return res;
}

