/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2023 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include "config.h"

#include <librnd/hid/hid.h>
#include <librnd/hid/hid_inlines.h>
#include "data.h"
#include "obj_any.h"

#include "search.h"

static osm_any_obj_t *osm_search_obj_xy_(osm_layer_t *ly, rnd_coord_t x, rnd_coord_t y, int unsel_only, osm_objtype_mask_t mask)
{
	osm_rtree_box_t box;
	int z;

	box.x1 = x - osm_slop; box.y1 = y - osm_slop;
	box.x2 = x + osm_slop; box.y2 = y + osm_slop;

	for(z = 0; z < OSM_ZORD_MAX; z++) {
		osm_any_obj_t *obj, *best_obj = NULL;
		osm_rtree_it_t it;
		void *o;

		for(o = osm_rtree_first(&it, &ly->zord[z], &box); o != NULL; o = osm_rtree_next(&it)) {
			obj = o;
			if (obj->hdr.type == OSM_OBJ_invalid) continue;
			if (unsel_only && obj->hdr.selected) continue;

			if (!obj->hdr.calls->isc_box(obj, &box)) continue;

			if (mask != 0) {
				osm_objtype_mask_t objm = 1 << obj->hdr.type;
				if ((mask & objm) == 0)
					continue;
			}

			if ((best_obj == NULL) || (obj->hdr.type < best_obj->hdr.type))
				best_obj = obj;
		}

		if (best_obj != NULL)
			return best_obj;
	}

	return NULL;
}

osm_any_obj_t *osm_search_obj_xy(osm_layer_t *ly, rnd_coord_t x, rnd_coord_t y)
{
	return osm_search_obj_xy_(ly, x, y, 0, 0);
}

osm_any_obj_t *osm_search_obj_xy_unsel(osm_layer_t *ly, rnd_coord_t x, rnd_coord_t y)
{
	return osm_search_obj_xy_(ly, x, y, 1, 0);
}

osm_any_obj_t *osm_search_obj_xy_mask(osm_layer_t *ly, rnd_coord_t x, rnd_coord_t y, osm_objtype_mask_t mask)
{
	return osm_search_obj_xy_(ly, x, y, 0, mask);
}

void osm_search_append_objs_box_mask(vtp0_t *dst, osm_layer_t *ly, const osm_rtree_box_t *box, osm_objtype_mask_t mask)
{
	int z;

	for(z = 0; z < OSM_ZORD_MAX; z++) {
		osm_rtree_it_t it;
		osm_any_obj_t *obj;
		void *o;

		for(o = osm_rtree_first(&it, &ly->zord[z], box); o != NULL; o = osm_rtree_next(&it)) {
			obj = o;
			if (obj->hdr.type == OSM_OBJ_invalid) continue;
			if (mask != 0) {
				osm_objtype_mask_t objm = 1 << obj->hdr.type;
				if ((mask & objm) == 0)
					continue;
			}
			if (!obj->hdr.calls->isc_box(obj, box))
				continue;
			vtp0_append(dst, obj);
		}
	}
}

void osm_search_append_objs_box(vtp0_t *dst, osm_layer_t *ly, const osm_rtree_box_t *box)
{
	osm_search_append_objs_box_mask(dst, ly, box, 0);
}

void osm_search_append_objs_xy_mask(vtp0_t *dst, osm_layer_t *ly, rnd_coord_t x, rnd_coord_t y, osm_objtype_mask_t mask)
{
	osm_rtree_box_t box;

	box.x1 = x - osm_slop; box.y1 = y - osm_slop;
	box.x2 = x + osm_slop; box.y2 = y + osm_slop;

	osm_search_append_objs_box_mask(dst, ly, &box, mask);
}

void osm_search_append_objs_xy(vtp0_t *dst, osm_layer_t *ly, rnd_coord_t x, rnd_coord_t y)
{
	osm_search_append_objs_xy_mask(dst, ly, x, y, 0);
}

long osm_search_all_selected_mask(vtp0_t *dst, osm_layer_t *ly, osm_objtype_mask_t mask)
{
	htIp_entry_t *e;
	int t;
	long cnt = 0;

	for(t = 0; t < OSM_OBJ_max; t++) {
		for(e = htIp_first(&ly->id2obj[t]); e != NULL; e = htIp_next(&ly->id2obj[t], e)) {
			osm_any_obj_t *obj = e->value;
			if (obj->hdr.selected) {
				if (mask != 0) {
					osm_objtype_mask_t objm = 1 << obj->hdr.type;
					if ((mask & objm) == 0)
						continue;
				}
				vtp0_append(dst, obj);
				cnt++;
			}
		}
	}

	return cnt;
}

long osm_search_all_selected(vtp0_t *dst, osm_layer_t *ly)
{
	return osm_search_all_selected_mask(dst, ly, 0);
}
