/*
 *                            COPYRIGHT
 *
 *  pcb-rnd, interactive printed circuit board design
 *  Copyright (C) 2019,2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://www.repo.hu/projects/librnd
 *    lead developer: http://www.repo.hu/projects/librnd/contact.html
 *    mailing list: (not yet)
 */

#include "config.h"

#include <stdio.h>
#include <ctype.h>

#include <librnd/core/rnd_printf.h>
#include <librnd/core/plugins.h>
#include <librnd/core/safe_fs.h>
#include <librnd/core/compat_misc.h>
#include <librnd/core/compat_fs.h>
#include <librnd/core/conf_multi.h>
#include "lib_wgetosm.h"

#include "lib_wgetosm_conf.h"

conf_lib_wgetosm_t conf_lib_wgetosm;

static const char lib_wgetosm_cookie[] = "libwget_osm";

#include "conf_internal.c"

static char *osm_wget_command(const char *url, const char *ofn, int update, const osm_wget_opts_t *opts, osm_wget_response_t *resp)
{
	gds_t tmp;

	gds_init(&tmp);
	rnd_append_printf(&tmp, "%s -O '%s'", conf_lib_wgetosm.plugins.lib_wgetosm.wget_cmd, ofn);
	if (update)
		gds_append_str(&tmp, " -c");

	if (resp != NULL)
		gds_append_str(&tmp, " -S");

	if (opts != NULL) {
		const char **h;
		if (opts->header != NULL) {
			for(h = opts->header; *h != NULL; h++) {
				gds_append_str(&tmp, " --header '");
				gds_append_str(&tmp, *h);
				gds_append(&tmp, '\'');
			}
		}
		if (opts->post_file != NULL) {
			gds_append_str(&tmp, " '--post-file=");
			gds_append_str(&tmp, opts->post_file);
			gds_append(&tmp, '\'');
		}
		if (opts->conf_file != NULL) {
			gds_append_str(&tmp, " '--config=");
			gds_append_str(&tmp, opts->conf_file);
			gds_append(&tmp, '\'');
		}
		if (opts->extra_args != NULL) {
			gds_append(&tmp, ' ');
			gds_append_str(&tmp, opts->extra_args);
		}
	}

	gds_append(&tmp, ' ');
	gds_append(&tmp, '\'');
	gds_append_str(&tmp, url);
	gds_append(&tmp, '\'');


	if (resp != NULL) {
		resp->tmpfn = rnd_tempfile_name_new("wget.hdr.txt");
		gds_append_str(&tmp, " 2>");
		gds_append_str(&tmp, resp->tmpfn);
	}

	return tmp.array;
}

int osm_wget_read_response(osm_wget_response_t *resp)
{
	FILE *f;
	char line[8192], *s, *sep, *end;
	int res = 0;

	if (resp == NULL)
		return 0;

	f = rnd_fopen(NULL, resp->tmpfn, "r");
	resp->status = -1;
	resp->error = NULL;

	if (f == NULL)
		return -1;

	while((s = fgets(line, sizeof(line), f)) != NULL) {

		/* header lines start with spaces */
		if (*s != ' ')
			continue;

		/* left-strip */
		while(isspace(*s)) s++;
		if (*s == '\0')
			continue;

		if (resp->status < 0) { /* first line */
			long l;

			sep = strchr(s, ' ');
			if (sep == NULL) {
				res = resp->status = -3;
				break;
			}

			sep++;
			l = strtol(sep, &end, 10);
			if ((*end != '\0') && (!isspace(*end))) {
				res = resp->status = -2;
				break;
			}
			resp->status = l;
			if ((l < 100) || (l > 999)) {
				res = -4;
				break;
			}
			continue;
		}

		/* subsequent lines */
		sep = strchr(s, ':');
		if (sep != NULL) {
			*sep = '\0';
			sep++;
			if (rnd_strcasecmp(s, "error") == 0)
				resp->error = rnd_strdup(sep);
		}
	}


	fclose(f);
	return res;
}


int osm_wget_disk_resp(const char *url, const char *ofn, int update, const osm_wget_opts_t *opts, osm_wget_response_t *resp)
{
	int res;
	char *cmd = osm_wget_command(url, ofn, update, opts, resp);

	res = rnd_system(NULL, cmd);
	free(cmd);

	osm_wget_read_response(resp);
	return res;
}

int osm_wget_disk(const char *url, const char *ofn, int update, const osm_wget_opts_t *opts)
{
	return osm_wget_disk_resp(url, ofn, update, opts, NULL);
}


FILE *osm_wget_popen_resp(const char *url, int update, const osm_wget_opts_t *opts, osm_wget_response_t *resp)
{
	FILE *f;
	char *cmd = osm_wget_command(url, "-", update, opts, resp);

	f = rnd_popen(NULL, cmd, "r");
	free(cmd);
	return f;
}

FILE *osm_wget_popen(const char *url, int update, const osm_wget_opts_t *opts)
{
	return osm_wget_popen_resp(url, update, opts, NULL);
}

void osm_wget_response_uninit(osm_wget_response_t *resp)
{
	free(resp->error); resp->error = NULL;
	rnd_tempfile_unlink(resp->tmpfn);
}


int pplg_check_ver_lib_wgetosm(int ver_needed) { return 0; }

void pplg_uninit_lib_wgetosm(void)
{
	rnd_conf_plug_unreg("plugins/lib_wgetosm/", lib_wgetosm_conf_internal, lib_wgetosm_cookie);
}

int pplg_init_lib_wgetosm(void)
{
	RND_API_CHK_VER;

	rnd_conf_plug_reg(conf_lib_wgetosm, lib_wgetosm_conf_internal, lib_wgetosm_cookie);

#define conf_reg(field,isarray,type_name,cpath,cname,desc,flags) \
	rnd_conf_reg_field(conf_lib_wgetosm, field,isarray,type_name,cpath,cname,desc,flags);
#include "lib_wgetosm_conf_fields.h"

	return 0;
}
