/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer - Optimizations
 *  Copyright (C) 2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

static void osm_opt_ortho_search(osm_layer_t *ly, vtp0_t *way_list, osm_node_t *dir[2])
{
	long n, m, npts;
	htIp_t wynd;
	vtp0_t all = {0};

	htIp_init(&wynd, htIp_keyhash, htIp_keyeq);

	osm_search_all_selected(&all, ly);
	for(n = 0; n < all.used; n++) {
		osm_way_t *way = all.array[n];

		if (way->hdr.type != OSM_OBJ_WAY)
			continue;

		/* remember all ways and waypoints */
		osm_way_ensure_ndp(way);
		for(m = 0; m < way->ndp.used; m++) {
			osm_node_t *nd = way->ndp.array[m];
			htIp_set(&wynd, nd->hdr.id, way);
		}
		vtp0_append(way_list, way);
	}

	/* find nodes that are not in ways */
	npts = 0;
	for(n = 0; n < all.used; n++) {
		osm_node_t *nd = all.array[n];
		if ((nd->hdr.type != OSM_OBJ_NODE) || htIp_has(&wynd, nd->hdr.id))
			continue;
		if (npts < 2)
			dir[npts] = nd;
		npts++;
	}

	if (npts != 2)
		dir[0] = dir[1] = NULL;

	htIp_uninit(&wynd);
	vtp0_uninit(&all);
}

static void osm_rot_way(osm_way_t *way, rnd_coord_t ox, rnd_coord_t oy, double cosa, double sina)
{
	long n;
	for(n = 0; n < way->ndp.used; n++) {
		osm_node_t *nd = way->ndp.array[n];
		if (nd->hdr.MARK)
			continue;
		rnd_rotate(&nd->x, &nd->y, ox, oy, cosa, sina);
		nd->hdr.MARK = 1;
	}
}

static void osm_unmark_way(osm_way_t *way)
{
	long n;
	for(n = 0; n < way->ndp.used; n++) {
		osm_node_t *nd = way->ndp.array[n];
		nd->hdr.MARK = 0;
		nd->hdr.tmp_grp[0] = NULL;
		nd->hdr.tmp_grp[1] = NULL;
	}
}

typedef enum grp_dir_e {
	DIR_HORIZ = 0,
	DIR_VERT = 1
} grp_dir_t;

typedef struct {
	vtp0_t nd; /* list of nodes in this group */
	grp_dir_t dir;
	long uid;
	gdl_elem_t link;
} grp_t;


/* Append nd to grp in direction dir, overriding nd's original grp membership */
static void grp_add(grp_dir_t dir, grp_t *grp, osm_node_t *nd)
{
	assert(grp->dir == dir);
	vtp0_append(&grp->nd, nd);
	nd->hdr.tmp_grp[dir] = grp;

	rnd_trace(" grp_add: G%c%ld += #" OSM_PID "\n",
		dir == DIR_HORIZ ? 'h':'v', grp->uid, nd->hdr.id);
}

static void grp_new(gdl_list_t *grps, grp_dir_t dir, osm_node_t *nd1, osm_node_t *nd2)
{
	static long uid;
	grp_t *grp = calloc(sizeof(grp_t), 1);
	grp->dir = dir;
	grp->uid = ++uid;

	rnd_trace(" grp_new: G%c%ld\n",
		dir == DIR_HORIZ ? 'h':'v', grp->uid);

	grp_add(dir, grp, nd1);
	grp_add(dir, grp, nd2);
	gdl_append(grps, grp, link);
}

static void grp_free(gdl_list_t *grps, grp_t *grp)
{
	rnd_trace(" grp_free: G%c%ld\n",
		grp->dir == DIR_HORIZ ? 'h':'v', grp->uid);

	gdl_remove(grps, grp, link);
	vtp0_uninit(&grp->nd);
	free(grp);
}

static void grp_merge(gdl_list_t *grps, grp_dir_t dir, grp_t *dst, grp_t *src)
{
	long n;

	rnd_trace(" grp_merge: G%c%ld += G%c%ld {\n",
		src->dir == DIR_HORIZ ? 'h':'v', src->uid,
		dst->dir == DIR_HORIZ ? 'h':'v', dst->uid);

	assert(dir == dst->dir);
	assert(dir == src->dir);

	for(n = 0; n < src->nd.used; n++)
		grp_add(dir, dst, src->nd.array[n]);

	grp_free(grps, src);
	rnd_trace("}\n");
}

static void grp_align(grp_t *grp)
{
	long n;
	double avg = 0;
	rnd_coord_t res;

	if (grp->nd.used <= 1)
		return;

	for(n = 0; n < grp->nd.used; n++) {
		osm_node_t *nd = grp->nd.array[n];
		switch(grp->dir) {
			case DIR_HORIZ: avg += nd->x; break;
			case DIR_VERT:  avg += nd->y; break;
		}
	}

	res = rnd_round(avg / grp->nd.used);

	for(n = 0; n < grp->nd.used; n++) {
		osm_node_t *nd = grp->nd.array[n];
		switch(grp->dir) {
			case DIR_HORIZ: nd->x = res; break;
			case DIR_VERT:  nd->y = res; break;
		}
	}
}

static int osm_ortho_way_map(gdl_list_t *grps, osm_way_t *way, long *fail)
{
	long n;
	int changed;

	for(n = 1; n < way->ndp.used; n++) {
		osm_node_t *nd1 = way->ndp.array[n-1];
		osm_node_t *nd2 = way->ndp.array[n];
		rnd_coord_t dx = nd2->x - nd1->x, dy = nd2->y - nd1->y;
		int dir = -1;

		if (dx < 0) dx = -dx;
		if (dy < 0) dy = -dy;

		if ((dx < 2) || (dy < 2)) /* already orthogonal enough */
			continue;

		if (dx < (dy*0.75)) { /* tries to be vertical */
			dir = DIR_HORIZ;
			changed=1;
		}
		else if (dy < (dx*0.75)) { /* tries to be horizontal */
			dir = DIR_VERT;
			changed=1;
		}
		else {
			rnd_trace("  fail %ld %ld\n", dx, dy);
			(*fail)++;
		}

		if (dir >= 0) {
			if ((nd1->hdr.tmp_grp[dir] != NULL) && (nd2->hdr.tmp_grp[dir] != NULL)) {
				if (nd1->hdr.tmp_grp[dir] != nd2->hdr.tmp_grp[dir])
					grp_merge(grps, dir, nd1->hdr.tmp_grp[dir], nd2->hdr.tmp_grp[dir]);
				/* else: they are already in the same group so we are fine */
			}
			else if (nd1->hdr.tmp_grp[dir])
				grp_add(dir, nd1->hdr.tmp_grp[dir], nd2);
			else if (nd2->hdr.tmp_grp[dir])
				grp_add(dir, nd2->hdr.tmp_grp[dir], nd1);
			else
				grp_new(grps, dir, nd1, nd2);
		}
	}

	return changed;
}

static void osm_ortho_way_exec(gdl_list_t *grps)
{
	grp_t *grp;

	while((grp = gdl_first(grps)) != NULL) {
		rnd_trace("grp_exec: G%c%ld\n",
			grp->dir == DIR_HORIZ ? 'h':'v', grp->uid);

		grp_align(grp);
		grp_free(grps, grp);
	}
}

static void osm_apply_way(osm_way_t *way)
{
	long n;

	for(n = 0; n < way->ndp.used; n++) {
		osm_node_t *nd = way->ndp.array[n];
		double lon, lat;

		if (nd->hdr.MARK)
			continue;

		rnd2lldeg(&lon, &lat, nd->x, nd->y);
		osm_node_modify(way->hdr.layer, nd, &lon, &lat, 1);
		nd->hdr.MARK = 1;
	}
}

typedef struct {
	double cx, cy, ang, ang_weight;
	long num_nodes;
} avgdir_t;

static void osm_avgdir_way(avgdir_t *ctx, osm_way_t *way)
{
	long n;

	for(n = 0; n < way->ndp.used; n++) {
		osm_node_t *nd1 = way->ndp.array[n];
		ctx->cx += nd1->x;
		ctx->cy += nd1->y;
		ctx->num_nodes++;
		if (n > 0) {
			osm_node_t *nd2 = way->ndp.array[n-1];
			double dx = nd2->x - nd1->x, dy = nd2->y - nd1->y, len;

			if ((dx > -5) && (dx < 5) && (dy > -5) && (dy < 5))
				continue; /* too short, ignore */

			/* make all vectors point up */
			if (dy < 0) {
				dy = -dy;
				dx = -dx;
			}
			if (dx < 0) /* ignore anything not in the first quarter */
				continue;
			
			len = sqrt(dx*dx + dy*dy);
			ctx->ang += atan2(dy, dx) * len;
			ctx->ang_weight += len;
		}
	}
}


static void osm_opt_ortho(osm_layer_t *ly, vtp0_t *ways, osm_node_t *dir[2])
{
	double ddx, ddy; /* default direction */
	double rotang, cosa, sina;
	rnd_coord_t ox, oy;
	long n, fail;
	gdl_list_t grps = {0};

	if (dir[0] == NULL) {
		avgdir_t ctx = {0};
		for(n = 0; n < ways->used; n++)
			osm_avgdir_way(&ctx, ways->array[n]);
		if (ctx.ang_weight == 0) {
			rnd_message(RND_MSG_ERROR, "Failed to find enough vectors pointing the right direction\n");
			return;
		}
		rotang = ctx.ang / ctx.ang_weight;
		ox = ctx.cx / ctx.num_nodes;
		oy = ctx.cy / ctx.num_nodes;
	}
	else {
		ddx = dir[1]->x - dir[0]->x;
		ddy = dir[1]->y - dir[0]->y;
		if ((ddx == 0) && (ddy == 0)) {
			rnd_message(RND_MSG_ERROR, "Directional nodes too close");
			return;
		}
		rotang = atan2(ddy, ddx);
		ox = dir[1]->x;
		oy = dir[1]->y;
	}

	/* clean initial state */
	for(n = 0; n < ways->used; n++)
		osm_unmark_way(ways->array[n]);

	/* rotate everything */
	cosa = cos(rotang);
	sina = sin(rotang);

	for(n = 0; n < ways->used; n++)
		osm_rot_way(ways->array[n], ox, oy, cosa, sina);
	for(n = 0; n < ways->used; n++)
		osm_unmark_way(ways->array[n]);

	/* do the actual transformation: map groups first then transform group by group */
	fail = 0;
	for(n = 0; n < ways->used; n++)
		osm_ortho_way_map(&grps, ways->array[n], &fail);
	osm_ortho_way_exec(&grps);

	/* rotate everything back */
	cosa = cos(-rotang);
	sina = sin(-rotang);

	for(n = 0; n < ways->used; n++)
		osm_rot_way(ways->array[n], ox, oy, cosa, sina);
	for(n = 0; n < ways->used; n++)
		osm_unmark_way(ways->array[n]);

	/* perform the actual modifications */
	uundo_freeze_serial(&ly->undo);
	for(n = 0; n < ways->used; n++)
		osm_apply_way(ways->array[n]);
	for(n = 0; n < ways->used; n++)
		osm_unmark_way(ways->array[n]);
	uundo_unfreeze_serial(&ly->undo);
	uundo_inc_serial(&ly->undo);

	if (fail != 0)
		rnd_message(RND_MSG_ERROR, "Failed to rotate %ld lines because they were too far off the generic direction\n", fail);
}
